<?php
// ----------------------------------------------------------------------------------
// Class: DbStore
// ----------------------------------------------------------------------------------
/**
 * DbStore is a persistent store of RDF data using relational database technology.
 * DbStore uses the MDB Library for PHP (http://pear.php.net/MDB),
 * which allows to connect to multiple databases in a portable manner.
 *
 * <BR><BR>History:<UL>
 * <LI>06-17-2003                : First version of this class</LI>
 * <LI>09-01-2003                : Port to MDB</LI>
 *
 * @version V0.7
 * @author Radoslaw Oldakowski <radol@gmx.de>
 * @package model
 * @access public
 */

class DbStore extends Object
{
    /**
     * Database connection object
     *
     * @var object ADOConnection
     * @access private
     */
    var $dbConn;

    /**
     * Constructor:
     * Set the database connection with the given parameters.
     *
     * @param string $dbDriver
     * @param string $host
     * @param string $dbName
     * @param string $user
     * @param string $password
     * @access public
     */
    function DbStore($dsn, $options = null)
    {
        require_once 'MDB.php';
        // create a new connection object
        $this->dbConn =& MDB::connect($dsn, $options);
        // connect to database
        // optimized for speed
        $this->dbConn->setFetchMode(MDB_FETCHMODE_ORDERED);
    }

    /**
     * Create tables and indexes for the given database type.
     * Currently supported: MsAccess and MySQL.
     * If you want to use other databases, you will have to create tables by yourself
     * according to the abstract <a href="database_schema.html">database schema</a>
     * described in the API documentation.
     *
     * @param string $databaseType
     * @throws PhpError
     * @access public
     */
    function createTables($databaseType)
    {
        if (strtolower($databaseType) == 'asaccess') {
            $this->_createTables_MsAccess();
        } elseif (strtolower($databaseType) == 'mysql') {
            $this->_createTables_MySql();
        } else {
            $errmsg = RDFAPI_ERROR . "(class: DbStore; method: createTables('$databaseType')):
                                  Currently only MsAcces and MySQL supported.";
            trigger_error($errmsg, E_USER_ERROR);
        }
    }

    /**
     * List all DbModels stored in the database.
     *
     * @return array
     * @throws SqlError
     * @access public
     */
    function listModels()
    {
        $sql = 'SELECT modelURI, baseURI FROM models';
        $result = $this->dbConn->queryAll($sql);
        if (MDB::isError($result)) {
            echo $result->getUserInfo();
        } else {
            for ($i=0,$j=count($result);$i<$j;++$i) {
                $models[$i]['modelURI'] = $result[$i][0];
                $models[$i]['baseURI'] = $result[$i][1];
            }
            return $models;
        }
    }

    /**
     * Check if the DbModel with the given modelURI is already stored in the database
     *
     * @param string $modelURI
     * @return boolean
     * @throws SqlError
     * @access public
     */
    function modelExists($modelURI)
    {
        $sql = "SELECT COUNT(*)
            FROM models
            WHERE modelURI = '{$modelURI}'";
        $result = $this->dbConn->queryOne($sql);
        if (MDB::isError($result)) {
            echo $result->getUserInfo();
        } else {
            if (!$result) {
                return false;
            }
            return true;
        }
    }

    /**
     * Create a new instance of DbModel with the given $modelURI and
     * load the corresponding values of modelID and baseURI from the database.
     * Return FALSE if the DbModel does not exist.
     *
     * @param string $modelURI
     * @return object DbModel
     * @access public
     */
    function getModel($modelURI)
    {
        if (!$this->modelExists($modelURI)) {
            return false;
        } else {
            $modelVars = $this->dbConn->queryRow("SELECT modelURI, modelID, baseURI FROM models WHERE modelURI='{$modelURI}'");

            return new DbModel($this->dbConn, $modelVars[0],
                $modelVars[1], $modelVars[2]);
        }
    }

    /**
     * Create a new instance of DbModel with the given $modelURI
     * and insert the DbModel variables into the database.
     * Return FALSE if there is already a model with the given URI.
     *
     * @param string $modelURI
     * @param string $baseURI
     * @return object DbModel
     * @throws SqlError
     * @access public
     */
    function getNewModel($modelURI, $baseURI = null)
    {
        if ($this->modelExists($modelURI)) {
            return false;
        } else {
            $modelID = $this->_createUniqueModelID();

            $result = &$this->dbConn->query("INSERT INTO models
                                            VALUES ('{$modelID}',
                                                    '{$modelURI}',
                                                    '{$baseURI}')");
            if (MDB::isError($result)) {
                echo $result->getUserInfo();
            } else {
                return new DbModel($this->dbConn, $modelURI, $modelID, $baseURI);
            }
        }
    }

    /**
     * Store a MemModel or another DbModel from a different DbStore in the database.
     * Return FALSE if there is already a model with modelURI matching the modelURI
     * of the given model.
     *
     * @param object $ Model  &$model
     * @param string $modelURI
     * @return boolean
     * @access public
     */
    function putModel(&$model, $modelURI = null)
    {
        if (!$modelURI) {
            if (is_a($model, 'MemModel')) {
                $modelURI = 'DbModel-' . $this->_createUniqueModelID();
            } else {
                $modelURI = $model->modelURI;
            }
        } else {
            if ($this->modelExists($modelURI)) {
                return false;
            }
        }

        $newDbModel = $this->getNewModel($modelURI, $model->getBaseURI());
        $newDbModel->addModel($model);
    }

    /**
     * Close the DbStore.
     * !!! Warning: If you close the DbStore all active instances of DbModel from this
     * !!!          DbStore will lose their database connection !!!
     *
     * @access public
     */
    function close()
    {
        $this->dbConn->disconnect();
        $this = null;
    }
    // =============================================================================
    // **************************** private methods ********************************
    // =============================================================================
    /**
     * Create a unique ID for the DbModel to be insert into the models table.
     * This method was implemented because some databases do not support auto-increment.
     *
     * @return integer
     * @access private
     */
    function _createUniqueModelID()
    {
        $maxModelID = $this->dbConn->queryOne('SELECT MAX(modelID) FROM models');
        return ++$maxModelID;
    }

    /**
     * Create tables and indexes for MsAccess database
     *
     * @throws SqlError
     * @access private
     */
    function _createTables_MsAccess()
    {
        $this->dbConn->autoCommit(false);

        $this->dbConn->query('CREATE TABLE models
                           (modelID long primary key,
                            modelURI varchar not null,
                            baseURI varchar)');

        $this->dbConn->query('CREATE UNIQUE INDEX m_modURI_idx ON models (modelURI)');

        $this->dbConn->query('CREATE TABLE statements
                           (modelID long,
                            subject varchar,
                            predicate varchar,
                            object Memo,
                            l_language varchar,
                            l_datatype varchar,
                            subject_is varchar(1),
                            object_is varchar(1),
                            primary key (modelID, subject, predicate, object,
                                         l_language, l_datatype))');

        $this->dbConn->query('CREATE INDEX s_mod_idx ON statements (modelID)');
        $this->dbConn->query('CREATE INDEX s_sub_idx ON statements (subject)');
        $this->dbConn->query('CREATE INDEX s_pred_idx ON statements (predicate)');
        $this->dbConn->query('CREATE INDEX s_obj_idx ON statements (object)');

        $result = $this->dbConn->commit();
        if (MDB::isError($result)) {
            echo $result->getUserInfo();
        }
        $this->dbConn->autoCommit(true);
    }

    /**
     * Create tables and indexes for MySQL database
     *
     * @throws SqlError
     * @access private
     */
    function _createTables_MySql()
    {
        $this->dbConn->autoCommit(false);

        $this->dbConn->query("CREATE TABLE models
                           (modelID bigint NOT NULL,
                            modelURI varchar(255) NOT NULL,
                            baseURI varchar(255) DEFAULT '',
                            primary key (modelID))"
        );

        $this->dbConn->query('CREATE UNIQUE INDEX m_modURI_idx ON models (modelURI)');

        $this->dbConn->query("CREATE TABLE statements
                           (modelID bigint NOT NULL,
                            subject varchar(255) NOT NULL,
                            predicate varchar(255) NOT NULL,
                            object text,
                            l_language varchar(255) DEFAULT '',
                            l_datatype varchar(255) DEFAULT '',
                            subject_is varchar(1) NOT NULL,
                            object_is varchar(1) NOT NULL)"
        );

        $this->dbConn->query('CREATE INDEX s_mod_idx ON statements (modelID)');
        $this->dbConn->query('CREATE INDEX s_sub_pred_idx ON statements
                          (subject(200),predicate(200))');
        $this->dbConn->qzery('CREATE INDEX s_obj_idx ON statements (object(250))');

        $result = $this->dbConn->commit();
        if (MDB::isError($result)) {
            echo $result->getUserInfo();
        }
        $this->dbConn->autoCommit(true);
    }
} // end: Class DbStore
?>