<?php
// ----------------------------------------------------------------------------------
// Class: Model
// ----------------------------------------------------------------------------------
/**
 * A model is a programming interface to an RDF graph.
 * An RDF graph is a directed labeled graph, as described in http://www.w3.org/TR/rdf-mt/.
 * It can be defined as a set of <S, P, O> triples, where P is a uriref, S is either
 * a uriref or a blank node, and O is either a uriref, a blank node, or a literal.
 *
 * <BR><BR>History:<UL>
 * <LI>11-13-2003                : Function saveAs added
 * <LI>07-27-2003                : This is an abstract parent class for MemModel and DbModel.
 *                                  The previous class Model has been renamed to MemModel
 *
 * @version V0.7
 * @author Radoslaw Oldakowski <radol@gmx.de>
 * @author Daniel Westphal <mail@d-westphal.de>
 * @package model
 * @access public
 */

Class Model extends Object
{
    /**
     * Base URI of the Model.
     * Affects creating of new resources and serialization syntax.
     *
     * @var string
     * @access private
     */
    var $baseURI;

    /**
     * Notice for people who are used to work with older versions of RAP.
     *
     * @throws PHPError
     * @access public
     */
    function Model()
    {
        $errmsg = 'Since RAP 0.6 the class for manipulating memory models has been renamed to MemModel.";
        $errmsg .= "<br>Sorry for this inconvenience.<br>';

        trigger_error($errmsg, E_USER_ERROR);
    }

    /**
     * Return current baseURI.
     *
     * @return string
     * @access public
     */
    function getBaseURI()
    {
        return $this->baseURI;
    }

    /**
     * Saves the RDF,N3 or N-Triple serialization of the MemModel to a file.
     * You can decide to which format the model should be serialized by using a
     * corresponding suffix-string as $type parameter. If no $type parameter
     * is placed this method will serialize the model to XML/RDF format.
     * Returns FALSE if the MemModel couldn't be saved to the file.
     *
     * @access public
     * @param string $filename
     * @param string $type
     * @throw PhpError
     * @return boolean
     */
    function saveAs($filename, $type = 'rdf')
    {
        // get suffix and create a corresponding serializer
        if ($type == 'rdf') {
            $ser = new RdfSerializer();
        } elseif ($type == 'nt') {
            $ser = new NTripleSerializer();
        } elseif ($type == 'n3') {
            $ser = new N3Serializer();
        } else {
            print ('Serializer type not properly defined. Use a string of "rdf","n3" or "nt".');
            return false;
        } ;

        return $ser->saveAs($this, $filename);
    }

    /**
     * Load a model from a file containing RDF, N3 or N-Triples.
     * This function recognizes the suffix of the filename (.n3 or .rdf) and
     * calls a suitable parser, if no $type is given as string ("rdf" "n3" "nt");
     * If the model is not empty, the contents of the file is added to this DbModel.
     *
     * @param string $filename
     * @param string $type
     * @access public
     */
    function load($filename, $type = null)
    {
        if ((isset($type)) && ($type == 'n3') OR ($type == 'nt')) {
            $parser = new N3Parser();
        } elseif ((isset($type)) && ($type == 'rdf')) {
            $parser = new RdfParser();
        } else {
            // create a parser according to the suffix of the filename
            // if there is no suffix assume the file to be XML/RDF
            preg_match("/\.([a-zA-Z0-9_]+)$/", $filename, $suffix);
            if (isset($suffix[1]) && (strtolower($suffix[1]) == 'n3')
                OR (strtolower($suffix[1]) == 'nt')
            ) {
                $parser = new N3Parser();
            } else {
                $parser = new RdfParser();
            }
        }

        $temp = &$parser->generateModel($filename);
        $this->addModel($temp);
        if ($this->getBaseURI() == null) {
            $this->setBaseURI($temp->getBaseURI());
        }
    }
} // end: Model

?>